#include <stdlib.h>
#include <stdbool.h>

#include "tsp.h"
#include "queue.h"
#include "lin_kernighan.h"


// Maximum recursion breadth for each recursion step
int breadth[] = {
		5, 2, 1, 1, 1, 1, 1, 1, 1, 1,
//		10, 5, 3, 2, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 0 };

int altBreadthA = 5;
int altBreadthB = 3;
int altBreadthD = 2;

/*int altBreadthA = 10;
int altBreadthB = 5;
int altBreadthD = 3;*/


// Neighbour lists
int nNeighbours = 0;
int** neighbours = NULL;

// The graph
Graph* g = NULL;


// Candidate list structures:
typedef enum CandidateType { CANDI_PREV, CANDI_NEXT } CandidateType;

typedef struct Candidate {
	int node;
	double rank;
	CandidateType type;
} Candidate;

// CandidateList is a kind of priority queue with limited length. Items rank is
// it's priority. First item has highest rank.
typedef struct CandidateList {
	int length;
	int maxLength;
	Candidate** order;
	
	Candidate* data;
} CandidateList;


/**
 * Get the distance between two nodes.
 */
inline double getDist(int a, int b) {
	return g->dists[a * g->n + b];
}

/**
 * Get gain in length when an edge between nodes baseNode and oldNeigh is 
 * replaced with a new edge between baseNode and newNeigh.
 */
inline double getChange(int baseNode, int oldNeigh, int newNeigh) {
	return getDist(baseNode, oldNeigh) - getDist(baseNode, newNeigh);
}


/**
 * Get the index of the next node on route.
 */
inline int next(Route* route, int a) {
	if (route->reverse)
		return route->order[(route->position[a] + route->n - 1) % route->n];
	else
		return route->order[(route->position[a] + 1) % route->n];
}


/**
 * Get the index of the previous node on route.
 */
inline int prev(Route* route, int a) {
	if (route->reverse)
		return route->order[(route->position[a] + 1) % route->n];
	else
		return route->order[(route->position[a] + route->n - 1) % route->n];
}

/**
 * Test if position k is between positions i and j on given route.
 * @return true, if i < k < j or k < j < i or j < i < k, false otherwise
 */
inline bool posSequence(Route* route, int i, int k, int j) {
	if (route->reverse)
		return (j < k && k < i) || (i < j && (k < i || j < k));
	else
		return (i < k && k < j) || (j < i && (k < j || i < k));
}

/**
 * Test if node b is between nodes a and c.
 */
inline bool sequence(Route* r, int a, int b, int c) {
	return posSequence(r, r->position[a], r->position[b], r->position[c]);
}


/**
 * Reverse the subsequence from a to b on given route.
 */
void flip(Route* route, int a, int b) {
	if (a == b) return;
	int i = route->position[a];
	int j = route->position[b];
	
	// reverse fix
	if (route->reverse) {
		int tmp = i; i = j; j = tmp;
	}
	
	// flip opposite sequence instead?
	bool longSeq = ((route->n + j - i + 1) % route->n) > (route->n / 2);
	if (longSeq) {
		int tmp = (i + route->n - 1) % route->n;
		i = (j + route->n + 1) % route->n;
		j = tmp;
		route->reverse = !(route->reverse);
	}
	
	// do the flip
	while (1) {
		int* a = &(route->order[i]);
		int* b = &(route->order[j]);
		
		// swap orders
		int tmp = *a; *a = *b; *b = tmp;
		
		// swap positions
		tmp = route->position[*a];
		route->position[*a] = route->position[*b];
		route->position[*b] = tmp;		
		
		// next pair
		i = (i + 1) % route->n;
		if (i == j) break;
		j = (j + route->n - 1) % route->n;
		if (i == j) break;
	}
}



/*Candidate* candidateData;
Candidate** candidatePtrs;
int nCandidatesAllocated;

void initCandidateSystem() {
	// calculate required number of candidates
	int n = 0;
	for (int i = 0; breadth[i] > 0; ++i)
		n += breadth[i] + 1;
	n += altBreadthA + 1;
	n += altBreadthB + 1;
	n += altBreadthD + 1;
	
	// allocate/initialize data and pointers
	candidateData = malloc(n * sizeof(Candidate));
	candidatePtrs = malloc(n * sizeof(Candidate*));
	for (int i = 0; i < n; ++i)
		candidatePtrs[i] = candidateData + i;
	
	nCandidatesAllocated = 0;
}

void freeCandidateSystem() {
	free(candidatePtrs);
	free(candidateData);
}*/


/**
 * Create/allocate a new candidate list of given length.
 */
CandidateList* createCandidateList(int maxLength) {
	CandidateList* cands = malloc(sizeof(CandidateList));
	cands->length = 0;
	cands->maxLength = maxLength;
	
	cands->data = malloc((maxLength + 1) * sizeof(Candidate));
	cands->order = malloc((maxLength + 1) * sizeof(Candidate*));
	for (int i = 0; i <= maxLength; ++i)
		cands->order[i] = cands->data + i;
	
	//cands->order = candidatePtrs + nCandidatesAllocated;
	//nCandidatesAllocated += maxLength + 1;
	
	return cands;
}

/**
 * Remove all elements from given candidate list.
 */
void resetCandidateList(CandidateList* cands) {
	cands->length = 0;
}

/**
 * Free candidate list.
 */
void freeCandidateList(CandidateList* cands) {
	free(cands->data);
	free(cands->order);
	free(cands);
	//nCandidatesAllocated -= cands->maxLength + 1;
}

/**
 * Insert a new item into candidate list. Calculates items rank value and places
 * it in to the right position in the list according to its rank. The higher the
 * rank, the smaller the position is in the list.
 * @param r      current route, used to calculate rank value
 * @param cands  the list the item is inserted
 * @param neigh  the node to be inserted
 * @param node   the node of which candidate list is being build, used to
 *               calculate rank value
 * @param type   item type: CANDI_PREV or CANDI_NEXT
 */
void insertIntoCandidateList(Route* r, CandidateList* cands, int neigh, int node,
		CandidateType type) {
	// construct new candidate
	Candidate* newCand = cands->order[cands->length];
	newCand->node = neigh;
	newCand->rank = getChange(neigh, type == CANDI_PREV ?
			prev(r, neigh) : next(r, neigh), node);
	newCand->type = type;
	
	// find the right position
	int i = cands->length;
	while (i > 0 && cands->order[i-1]->rank < newCand->rank) {
		cands->order[i] = cands->order[i-1];
		--i;
	}
	
	// insert
	cands->order[i] = newCand;
	if (cands->length < cands->maxLength)
		++(cands->length);
}



/**
 * Does one search step (involving one flip). Correspond to Algorithm 2.1 in the
 * report "Finding Tours in the TSP" (Applegate, Bixby, Chvatal, Cook)
 * (without Mak-Morton move)
 * @param r      route to optimize
 * @param base   base node of flips
 * @param level  recurseion depth
 * @param delta  the gain in route length this far
 * @return true if a better route was found, false otherwise
 */
bool lkStep(Route* r, int base, int level, double delta) {
	
	// breadth of search on this level
	int br = breadth[level];
	
	// maximum depth?
	if (br == 0)
		return false;
	
	int nextBase = next(r, base);
	
	// create candidate list
	CandidateList* candidates = createCandidateList(br);
	
	// maximum allowed neighbour distance and position limits
	double maxDist = delta + getDist(base, nextBase);
	int floorNode = next(r, nextBase);
	int ceilNode = base;
	
	// select best candidates from neighbours
	for (int i = 0; i < nNeighbours; ++i) {
		int neigh = neighbours[nextBase][i];
		
		// too far?
		if (getDist(nextBase, neigh) >= maxDist)
			break;
		
		// if not too close, try to insert into the candidate list
		if (sequence(r, floorNode, neigh, ceilNode))
			insertIntoCandidateList(r, candidates, neigh, nextBase, CANDI_PREV);
	}
	
	// test each canditate
	for (int i = 0; i < candidates->length; ++i) {
		int a = candidates->order[i]->node;
		int prevA = prev(r, a);
		
		// get improvement
		double deltaChange = getChange(nextBase, base, a) + getChange(prevA, a, base);
		
		// do flip
		flip(r, nextBase, prevA);
		
		// if flip sequence this far improves route => return
		if (delta + deltaChange > 0) {
			freeCandidateList(candidates);
			return true;
		}
		
		// continue flips
		if (lkStep(r, base, level + 1, delta + deltaChange)) {
			freeCandidateList(candidates);
			return true;
		}
		
		// if couldn't find improvement => undo flip
		flip(r, prevA, nextBase);
	}
	
	freeCandidateList(candidates);
	
	// no better route found
	return false;
}


/**
 * Does an alternate firtst search step (involving two or three flips).
 * Correspond to Algorithm 2.2 in the report "Finding Tours in the TSP"
 * (Applegate, Bixby, Chvatal, Cook)
 * @param r      route to optimize
 * @param base   base node of flips
 * @param level  recurseion depth
 * @param delta  the gain in route length this far
 * @return true if a better route was found, false otherwise
 */
bool lkAltStep(Route* r, int base, int level, double delta) {
	
	// create candidate lists
	CandidateList* candidatesA = createCandidateList(altBreadthA);
	CandidateList* candidatesB = createCandidateList(altBreadthB);
	CandidateList* candidatesD = createCandidateList(altBreadthD);
	
	int nextBase = next(r, base);
	
	// maximum allowed neighbour distance and position limits
	double maxDist = delta + getDist(base, nextBase);
	int floorNodeA = next(r, next(r, nextBase));
	int ceilNodeA = prev(r, base);
	
	//bool foundImprovedRoute = false;
	
	// select best candidates from neighbours of nextBase
	for (int i = 0; i < nNeighbours; ++i) {
		int neigh = neighbours[nextBase][i];
		
		// too far?
		if (getDist(nextBase, neigh) >= maxDist)
			break;
		
		// if not too close, try to insert into the candidate list
		if (sequence(r, floorNodeA, neigh, ceilNodeA))
			insertIntoCandidateList(r, candidatesA, neigh, nextBase, CANDI_NEXT);
	}
	
	// test each canditate A
	for (int i = 0; i < candidatesA->length; ++i) {
		int a = candidatesA->order[i]->node;
		
		int nextA = next(r, a);
		
		// reset candidate B list
		resetCandidateList(candidatesB);
		
		// maximum allowed neighbour distance and position limits
		double maxDist = getDist(a, nextA) + getChange(nextBase, base, a);
		int floorNodeB1 = nextBase;
		int ceilNodeB1 = a;
		int floorNodeB2 = next(r, nextA);
		int ceilNodeB2 = base;
		
		// select best candidates from neighbours of nextA
		for (int i = 0; i < nNeighbours; ++i) {
			int neigh = neighbours[nextA][i];
			
			// too far?
			if (getDist(nextA, neigh) >= maxDist)
				break;
			
			// case1: between nextBase and a
			if (sequence(r, floorNodeB1, neigh, ceilNodeB1)) {
				if (next(r, nextBase) != neigh)
					insertIntoCandidateList(r, candidatesB, neigh, nextA, CANDI_PREV);
				if (prev(r, a) != neigh)
					insertIntoCandidateList(r, candidatesB, neigh, nextA, CANDI_NEXT);
			}
			// case2: between nextA+1 and base
			else if (sequence(r, floorNodeB2, neigh, ceilNodeB2)) {
				insertIntoCandidateList(r, candidatesB, neigh, nextA, CANDI_PREV);
			}
		}
		
		// test each canditate B
		for (int i = 0; i < candidatesB->length; ++i) {
			int b = candidatesB->order[i]->node;
			
			// case 1: between nextBase and a
			if (sequence(r, nextBase, b, a)) {
				
				// CANDI_PREV
				if (candidatesB->order[i]->type == CANDI_PREV) {
					int prevB = prev(r, b);
					
					// get improvement
					double deltaChange = getChange(nextBase, base, a) +
							getChange(nextA, a, b) + getChange(prevB, b, base);
					
					// do flips
					flip(r, nextBase, prevB);
					flip(r, b, a);
					
					// if flip sequence this far improves route => return
					if (delta + deltaChange > 0) {
						freeCandidateList(candidatesD);
						freeCandidateList(candidatesB);
						freeCandidateList(candidatesA);
						return true;
					}
					
					// continue flips
					if (lkStep(r, base, level + 2, delta + deltaChange)) {
						freeCandidateList(candidatesD);
						freeCandidateList(candidatesB);
						freeCandidateList(candidatesA);
						return true;
					}
					
					// if couldn't find improvement => undo flips
					flip(r, a, b);
					flip(r, prevB, nextBase);
				}
				
				// CANDI_NEXT
				else {
					int nextB = next(r, b);
					
					// get improvement
					double deltaChange = getChange(nextBase, base, a) +
							getChange(nextA, a, b) + getChange(nextB, b, base);
					
					// do flips
					flip(r, nextBase, b);
					flip(r, nextB, a);
					flip(r, b, nextB);
					
					// if flip sequence this far improves route => return
					if (delta + deltaChange > 0) {
						freeCandidateList(candidatesD);
						freeCandidateList(candidatesB);
						freeCandidateList(candidatesA);
						return true;
					}
					
					// continue flips
					if (lkStep(r, base, level + 2, delta + deltaChange)) {
						freeCandidateList(candidatesD);
						freeCandidateList(candidatesB);
						freeCandidateList(candidatesA);
						return true;
					}
					
					// if couldn't find improvement => undo flips
					flip(r, nextB, b);
					flip(r, a, nextB);
					flip(r, b, nextBase);
				}
			}
			
			// case2: between nextA+1 and base
			else {
			
				int prevB = prev(r, b);
				
				// reset candidate D list
				resetCandidateList(candidatesD);
				
				// maximum allowed neighbour distance and position limits
				double maxDist = getDist(b, prevB) + getChange(nextBase, base, a) +
						getChange(nextA, a, b);
				int floorNodeD = floorNodeB1;
				int ceilNodeD = ceilNodeB1;
				
				// select best candidates from neighbours of prevB
				for (int i = 0; i < nNeighbours; ++i) {
					int neigh = neighbours[prevB][i];
			
					// too far?
					if (getDist(prevB, neigh) >= maxDist)
						break;
			
					// if between nextBase and a
					if (sequence(r, floorNodeD, neigh, ceilNodeD)) {
						if (next(r, nextBase) != neigh)
							insertIntoCandidateList(r, candidatesD, neigh, prevB, CANDI_PREV);
						if (prev(r, a) != neigh)
							insertIntoCandidateList(r, candidatesD, neigh, prevB, CANDI_NEXT);
					}
				}
				
				
				// test each canditate D
				for (int i = 0; i < candidatesD->length; ++i) {
					int d = candidatesD->order[i]->node;
					
					// CANDI_PREV
					if (candidatesD->order[i]->type == CANDI_PREV) {
						int prevD = prev(r, d);
					
						// get improvement
						double deltaChange = getChange(nextBase, base, a) +
								getChange(nextA, a, b) + getChange(prevB, b, d) +
								getChange(prevD, d, base);
						
						// do flips
						flip(r, nextBase, prevD);
						flip(r, d, a);
						flip(r, nextA, prevB);
						
						// if flip sequence this far improves route => return
						if (delta + deltaChange > 0) {
							freeCandidateList(candidatesD);
							freeCandidateList(candidatesB);
							freeCandidateList(candidatesA);
							return true;
						}
						
						// continue flips
						if (lkStep(r, base, level + 3, delta + deltaChange)) {
							freeCandidateList(candidatesD);
							freeCandidateList(candidatesB);
							freeCandidateList(candidatesA);
							return true;
						}
					
						// if couldn't find improvement => undo flips
						flip(r, prevB, nextA);
						flip(r, a, d);
						flip(r, prevD, nextBase);
					}
				
					// CANDI_NEXT
					else {
						int nextD = next(r, d);
					
						// get improvement
						double deltaChange = getChange(nextBase, base, a) +
								getChange(nextA, a, b) + getChange(prevB, b, d) +
								getChange(nextD, d, base);
						
						// do flips
						flip(r, nextBase, prevB);
						flip(r, prevB, nextD);
						flip(r, nextA, nextBase);
					
						// if flip sequence this far improves route => return
						if (delta + deltaChange > 0) {
							freeCandidateList(candidatesD);
							freeCandidateList(candidatesB);
							freeCandidateList(candidatesA);
							return true;
						}
					
						// continue flips
						if (lkStep(r, base, level + 3, delta + deltaChange)) {
							freeCandidateList(candidatesD);
							freeCandidateList(candidatesB);
							freeCandidateList(candidatesA);
							return true;
						}
					
						// if couldn't find improvement => undo flips
						flip(r, nextBase, nextA);
						flip(r, nextD, prevB);
						flip(r, prevB, nextBase);
					}
				}
			}
		}
	}
	
	freeCandidateList(candidatesD);
	freeCandidateList(candidatesB);
	freeCandidateList(candidatesA);
	
	// no better route found
	return false;
}




/**
 * Search for improving flip sequence. Correspond to Algorithm 2.3 in the
 * report "Finding Tours in the TSP" (Applegate, Bixby, Chvatal, Cook)
 * @param r      route to optimize
 * @param base   base node for search
 * @return true if a better route was found, false otherwise
 */
bool lkSearch(Route* r, int base) {
	if (lkStep(r, base, 0, 0)) {
		//printf("step improved\n");
		return true;
	} else if (lkAltStep(r, base, 0, 0)) {
		//printf("alternateStep improved\n");
		return true;
	}
	return false;
}



/**
 * Optimizes given route using Lin-Kernighan algorithm. Correspond to Algorithm
 * 2.4 in the report "Finding Tours in the TSP" (Applegate, Bixby, Chvatal, Cook)
 * @param r         route to optimize
 * @param callback  callback function, that is called after each improvement
 *                  if NULL, no function call is made
 */
void linKernighan(Route* r, void (*callback)(Route* route)) {
	// create ban-list
	bool banned[g->n];
	Queue* available = createQueue();
	for (int i = 0; i < g->n; ++i) {
		banned[i] = false;
		enqueue(available, i);
	}
	
	// call lkSearch using different base nodes until all nodes are banned
	while (queueSize(available) > 0) {
		int base = dequeue(available);
		banned[base] = true;
		if (lkSearch(r, base)) {
			// if improvement was found then make all nodes again available
			for (int i = 0; i < g->n; ++i)
				if (banned[i]) {
					banned[i] = false;
					enqueue(available, i);
				}
			
			// call given callback
			if (callback)
				callback(r);
		}
	}
	
	freeQueue(available);
}


/**
 * Prepare linKernighan to be ready to run.
 * @param graph    TSP graph
 * @param nNeighs  number of neighbours for each node
 * @param neighs   neighbours for each node
 */
void initLinKernighan(Graph* graph, int nNeighs, int** neighs) {
	g = graph;
	neighbours = neighs;
	nNeighbours = nNeighs;
}




