#include <stdio.h>
#include <stdbool.h>
//#include <unistd.h>
#include <getopt.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>

#include "tsp.h"
#include "tsp_loader.h"
#include "visualize.h"
//#include "queue.h"
#include "lin_kernighan.h"


int NUM_NEIGHBOURS_DEFAULT = 10;

void printRoute(Route* route) {
	if (route->reverse)
		for (int i = route->n - 1; i >= 0 ; --i)
			printf("%d, ", route->order[i]);
	else
		for (int i = 0; i < route->n; ++i)
			printf("%d, ", route->order[i]);
	printf("\b\b\n");
}

// Return the distance between two nodes.
inline double getDistance(Graph* g, int a, int b) {
	return g->dists[a * g->n + b];
}

// Calculates the length of given route.
double routeLength(Graph* g, Route* route) {
	int l = getDistance(g, route->order[route->n-1], route->order[0]);
	for (int i = 1; i < route->n; ++i)
		l += getDistance(g, route->order[i-1], route->order[i]);
	return l;
}

// Returns an array containing n closest neighbours for every node.
int** findClosestNeighbours(Graph* g, int nNeighbours) {
	int** neighbours = malloc(g->n * sizeof(int));
	for (int i = 0; i < g->n; ++i) {
		neighbours[i] = malloc((g->n + 1) * sizeof(int));
		for (int k = 0; k < nNeighbours; ++k)
			neighbours[i][k] = -1;
		for (int j = 0; j < g->n; ++j) {
			if (j == i)
				continue;
			
			int k = nNeighbours - 1;
			while (k >= 0 && (neighbours[i][k] == -1 ||
					g->dists[i * g->n + j] < g->dists[i * g->n + neighbours[i][k]])) {
				neighbours[i][k+1] = neighbours[i][k];
				neighbours[i][k] = j;
				--k;
			}
		}
	}
	return neighbours;
}

// Create a randomly selected route for n nodes.
Route* createRandomRoute(int n) {
	Route* r = malloc(sizeof(Route));
	r->n = n;
	r->order = malloc(n * sizeof(int));
	r->position = malloc(n * sizeof(int));
	r->reverse = false;
	
	for (int i = 0; i < n; ++i)
		r->order[i] = i;

	for (int i = 0; i < n; ++i) {
		int j = i + rand() % (n - i);

		int k = r->order[j];
		r->order[j] = r->order[i];
		r->order[i] = k;
		
		r->position[k] = i;
	}
	return r;
}

// Create route 1, 2, ..., n
Route* createIdentityRoute(int n) {
	Route* r = malloc(sizeof(Route));
	r->n = n;
	r->order = malloc(n * sizeof(int));
	r->position = malloc(n * sizeof(int));
	r->reverse = false;
	
	for (int i = 0; i < n; ++i) {
		r->order[i] = i;
		r->position[i] = i;
	}

	return r;
}

// Free/deallocate given route
void freeRoute(Route* r) {
	if (r) {
		if (r->order)
			free(r->order);
		if (r->position)
			free(r->position);
		free(r);
	}
}

// Draw current route in visualizer.
void drawState(Route* route) {
	clearPicture();
	drawRoute(route);
	drawNodes();
	updatePicture();
}


// Main
int main(int argc, char **argv) {
	// default options
	bool visualize = false;
	bool animate = false;
	bool initRand = true;
	bool iterate = false;
	int timeLimit = INT_MAX;
	bool quiet = false;
	bool printOptimizedRoute = false;
	int nNeighbours = NUM_NEIGHBOURS_DEFAULT;
	int picWidth = 800;
	int picHeight = 800;
	
	// parse options
	char c;
	while ((c = getopt(argc, argv, "hin:pqrs:t:v")) != -1)
		switch (c) {
			case 'h':
				printf(
					"Usage:\n"
					"    %s [options] filename\n"
					"\n"
					"Options:\n"
					"    -h        Print this help message.\n"
					"    -i        Iterative mode.\n"
					"    -n num    Set the number of neighbours in neighbour list.\n"
					"    -p        Print achieved route.\n"
					"    -q        Quiet mode.\n"
					"    -r        Disable random generator initializing.\n"
					"    -s WxH    Set visualizer window width to W and height to H.\n"
					"              Example: -s 800x600\n"
					"    -t sec    Time limit in seconds for iterations. Implies iterative mode.\n"
					"    -v        Use visualizer. Use animation if the option is given twice.\n"
					"              In visualizer, press any key to start the search, to stop\n"
					"              the search in animation mode and to exit (in this order).\n"
					, argv[0]);
				return 0;
			case 'i':
				iterate = true;
				break;
			case 'n':
				nNeighbours = atoi(optarg);
				if (nNeighbours <= 0) {
					fprintf(stderr, "Neighbour number must be a positive integer.\n");
					return 1;
				}
				break;
			case 'p':
				printOptimizedRoute = true;
				break;
			case 'q':
				quiet = true;
				break;
			case 'r':
				initRand = false;
				break;
			case 's':
				picWidth = atoi(optarg);
				if (picWidth < 100) {
					fprintf(stderr, "Picture size should be format WxH where W and H are integers (>100).\n");
					return 1;
				}
				char* sep = strchr(optarg, 'x');
				if (!sep) {
					fprintf(stderr, "Picture size should be format WxH where W and H are integers (>100).\n");
					return 1;
				}
				picHeight = atoi(sep + 1);
				if (picHeight < 100) {
					fprintf(stderr, "Picture size should be format WxH where W and H are integers (>100).\n");
					return 1;
				}
				break;
			case 't':
				iterate = true;
				timeLimit = atoi(optarg);
				if (timeLimit <= 0) {
					fprintf(stderr, "Time limit must be a positive integer.\n");
					return 1;
				}
				break;
			case 'v':
				if (visualize)
					animate = true;
				visualize = true;
				break;
			case '?':
				/*if (optopt == 'n')
					fprintf (stderr, "Option -%c requires an argument.\n", optopt);
				else if (isprint (optopt))
					fprintf(stderr, "Unknown option `-%c'.\n", optopt);
				else
					fprintf(stderr, "Unknown option character `\\x%x'.\n", optopt);*/
				return 1;
			default:
				abort();
		}
	
	// initialize random generator
	if (initRand)
		srand(time(NULL));
	
	// get filename
	if (optind == argc) {
		fprintf (stderr, "Filename required.\n");
		fprintf (stderr, "Type %s -h for help.\n", argv[0]);
		return 1;
	}
	if (optind < argc - 1) {
		fprintf (stderr, "Only one filename accepted.\n");
		fprintf (stderr, "Type %s -h for help.\n", argv[0]);
		return 1;
	}
	char* tspFilename = argv[optind];
	
	// load file
	TSP* tsp = loadTSP(tspFilename);
	if (!tsp)
		return 1;
	
	Graph* graph = malloc(sizeof(Graph));
	graph->n = tsp->nCities;
	graph->dists = tsp->dists;
	
	// find neighbours
	int** neighbours = findClosestNeighbours(graph, nNeighbours);
	
	// initialize linKernighan
	initLinKernighan(graph, nNeighbours, neighbours);
	
	// create initial route
	Route* route = createRandomRoute(tsp->nCities);
	
	if (visualize) {
		initVisualizer(picWidth, picHeight, 32, 0);
		setVisualizerTSP(tsp);
		drawState(route);
		waitKey();
	}
	
	clock_t startTime = clock();
	
	// optimize
	if (animate)
		linKernighan(route, drawState);
	else
		linKernighan(route, NULL);
	
	// iterate optimization
	int i = 1;
	while (iterate && (clock() - startTime) / CLOCKS_PER_SEC < timeLimit &&
			(!visualize || !keyPressed())) {
		// display current route and status
		if (visualize) {
			drawState(route);
			char caption[100];
			sprintf(caption, "Lin-Kernighan TSP solver: %s (length: %5.0lf) [iterations: %d]",
					tsp->name, routeLength(graph, route), i);
			SDL_WM_SetCaption(caption, NULL);
		}
		
		// create a new random initial route and optimize
		Route* newRoute = createRandomRoute(tsp->nCities);
		linKernighan(newRoute, NULL);
		
		// update current best route
		if (routeLength(graph, newRoute) < routeLength(graph, route)) {
			freeRoute(route);
			route = newRoute;
		} else {
			freeRoute(newRoute);
		}
		
		++i;
	}
	
	clock_t endTime = clock();
	
	// draw achieved route
	if (visualize) {
		drawState(route);
		char caption[100];
		sprintf(caption, "Lin-Kernighan TSP solver: %s (length: %5.0lf)", tsp->name, routeLength(graph, route));
		SDL_WM_SetCaption(caption, NULL);
		waitKey();
		quitVisualizer();
	}
	
	if (printOptimizedRoute) {
		printf("Generated route:\n");
		printRoute(route);
	}
	
	if (!quiet) {
		printf("Iterations: %d\n", i);
		printf("Time used: %.2f s\n", (endTime - startTime) / (float)CLOCKS_PER_SEC);
		printf("Route length: %.1f\n", routeLength(graph, route));
	}
	
	freeRoute(route);
	
	return 0;
}


