#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdbool.h>
#include <math.h>

#include "tsp.h"
#include "tsp_loader.h"

// some constants
const int MAX_KEYWORD_LENGTH = 30;
const int MAX_STRING_LENGTH = 500;

const double EARTH_RADIUS = 6378.388;
const double PI = 3.141592;


// keywords
typedef enum Keyword {
	KW_NOT_SET = 0,
	
	// main keywords
	KW_NAME,
	KW_TYPE,
	KW_COMMENT,
	KW_DIMENSION,
	KW_CAPACITY,
	KW_EDGE_WEIGHT_TYPE,
	KW_EDGE_WEIGHT_FORMAT,
	KW_EDGE_DATA_FORMAT,
	KW_NODE_COORD_TYPE,
	KW_DISPLAY_DATA_TYPE,
	
	KW_NODE_COORD_SECTION,
	KW_DEPOT_SECTION,
	KW_DEMAND_SECTION,
	KW_EDGE_DATA_SECTION,
	KW_FIXED_EDGES_SECTION,
	KW_DISPLAY_DATA_SECTION,
	KW_TOUR_SECTION,
	KW_EDGE_WEIGHT_SECTION,

	KW_EOF,
	
	// problem type
	KW_TSP,
	KW_ATSP,
	KW_SOP,
	KW_HCP,
	KW_CVRP,
	KW_TOUR,
	
	// edge weight type
	KW_EXPLICIT,
	KW_EUC_2D,
	KW_EUC_3D,
	KW_MAX_2D,
	KW_MAX_3D,
	KW_MAN_2D,
	KW_MAN_3D,
	KW_CEIL_2D,
	KW_GEO,
	KW_ATT,
	KW_XRAY1,
	KW_XRAY2,
	KW_SPECIAL,
	
	// edge weight format
	KW_FUNCTION,
	KW_FULL_MATRIX,
	KW_UPPER_ROW,
	KW_LOWER_ROW,
	KW_UPPER_DIAG_ROW,
	KW_LOWER_DIAG_ROW,
	KW_UPPER_COL,
	KW_LOWER_COL,
	KW_UPPER_DIAG_COL,
	KW_LOWER_DIAG_COL,
	
	// edge data format
	KW_EDGE_LIST,
	KW_ADJ_LIST,
	
	// node coord type
	KW_TWOD_COORDS,
	KW_THREED_COORDS,
	KW_NO_COORDS,
	
	// display data type
	KW_COORD_DISPLAY,
	KW_TWOD_DISPLAY,
	KW_NO_DISPLAY,
	
	
	KW_END_OF_KEYWORDS,
	KW_UNKNOWN
} Keyword;


// struct to connect keyword with its string representation
typedef struct KeywordStr {
	Keyword kw;
	char* str;
} KeywordStr;

// macros to define KeywordStr-arrays more easily
#define START_KEYWORD_STR_ARRAY(name) KeywordStr name[] = {
#define ADD_KEYWORD_STR(keyword) {KW_##keyword, #keyword},
#define END_KEYWORD_STR_ARRAY() {KW_END_OF_KEYWORDS, NULL} };


// main keywords
START_KEYWORD_STR_ARRAY(mainKeywords)
	ADD_KEYWORD_STR(NAME)
	ADD_KEYWORD_STR(TYPE)
	ADD_KEYWORD_STR(COMMENT)
	ADD_KEYWORD_STR(DIMENSION)
	ADD_KEYWORD_STR(CAPACITY)
	ADD_KEYWORD_STR(EDGE_WEIGHT_TYPE)
	ADD_KEYWORD_STR(EDGE_WEIGHT_FORMAT)
	ADD_KEYWORD_STR(EDGE_DATA_FORMAT)
	ADD_KEYWORD_STR(NODE_COORD_TYPE)
	ADD_KEYWORD_STR(DISPLAY_DATA_TYPE)
	
	ADD_KEYWORD_STR(NODE_COORD_SECTION)
	ADD_KEYWORD_STR(DEPOT_SECTION)
	ADD_KEYWORD_STR(DEMAND_SECTION)
	ADD_KEYWORD_STR(EDGE_DATA_SECTION)
	ADD_KEYWORD_STR(FIXED_EDGES_SECTION)
	ADD_KEYWORD_STR(DISPLAY_DATA_SECTION)
	ADD_KEYWORD_STR(TOUR_SECTION)
	ADD_KEYWORD_STR(EDGE_WEIGHT_SECTION)
	
	ADD_KEYWORD_STR(EOF)
END_KEYWORD_STR_ARRAY()

// problem type
START_KEYWORD_STR_ARRAY(problemTypeKeywords)
	ADD_KEYWORD_STR(TSP)
	ADD_KEYWORD_STR(ATSP)
	ADD_KEYWORD_STR(SOP)
	ADD_KEYWORD_STR(HCP)
	ADD_KEYWORD_STR(CVRP)
	ADD_KEYWORD_STR(TOUR)
END_KEYWORD_STR_ARRAY()

// edge weight type
START_KEYWORD_STR_ARRAY(edgeWeightTypeKeywords)
	ADD_KEYWORD_STR(EXPLICIT)
	ADD_KEYWORD_STR(EUC_2D)
	ADD_KEYWORD_STR(EUC_3D)
	ADD_KEYWORD_STR(MAX_2D)
	ADD_KEYWORD_STR(MAX_3D)
	ADD_KEYWORD_STR(MAN_2D)
	ADD_KEYWORD_STR(MAN_3D)
	ADD_KEYWORD_STR(CEIL_2D)
	ADD_KEYWORD_STR(GEO)
	ADD_KEYWORD_STR(ATT)
	ADD_KEYWORD_STR(XRAY1)
	ADD_KEYWORD_STR(XRAY2)
	ADD_KEYWORD_STR(SPECIAL)
END_KEYWORD_STR_ARRAY()

// edge weight format
START_KEYWORD_STR_ARRAY(edgeWeightFormatKeywords)
	ADD_KEYWORD_STR(FUNCTION)
	ADD_KEYWORD_STR(FULL_MATRIX)
	ADD_KEYWORD_STR(UPPER_ROW)
	ADD_KEYWORD_STR(LOWER_ROW)
	ADD_KEYWORD_STR(UPPER_DIAG_ROW)
	ADD_KEYWORD_STR(LOWER_DIAG_ROW)
	ADD_KEYWORD_STR(UPPER_COL)
	ADD_KEYWORD_STR(LOWER_COL)
	ADD_KEYWORD_STR(UPPER_DIAG_COL)
	ADD_KEYWORD_STR(LOWER_DIAG_COL)
END_KEYWORD_STR_ARRAY()

// edge data format
START_KEYWORD_STR_ARRAY(edgeDataFormatKeywords)
	ADD_KEYWORD_STR(EDGE_LIST)
	ADD_KEYWORD_STR(ADJ_LIST)
END_KEYWORD_STR_ARRAY()

// node coord type
START_KEYWORD_STR_ARRAY(nodeCoordTypeKeywords)
	ADD_KEYWORD_STR(TWOD_COORDS)
	ADD_KEYWORD_STR(THREED_COORDS)
	ADD_KEYWORD_STR(NO_COORDS)
END_KEYWORD_STR_ARRAY()

// display data type
START_KEYWORD_STR_ARRAY(displayDataTypeKeywords)
	ADD_KEYWORD_STR(COORD_DISPLAY)
	ADD_KEYWORD_STR(TWOD_DISPLAY)
	ADD_KEYWORD_STR(NO_DISPLAY)
END_KEYWORD_STR_ARRAY()


/**
 * Get the square of number.
 */
inline double sq(double x) {
	return x * x;
}

/**
 * Read one keyword from file. 
 * @param f                file descriptor
 * @param possibleKeywords array of allowed keywords
 * @return read keyword in success, KW_UNKNOWN otherwise
 */
Keyword readKeyword(FILE* f, KeywordStr const* possibleKeywords) {
	char kwBuffer[MAX_KEYWORD_LENGTH + 1];
	
	// read the string
	int l = 0;
	char c = getc(f);
	while (isalnum(c) || c == '_') {
		if (l < MAX_KEYWORD_LENGTH) {
			kwBuffer[l] = c;
			++l;
		}
		c = getc(f);
	}
	ungetc(c, f);
	kwBuffer[l] = '\0';
	//printf("*** Keyword: %s (%d) (eof: %d)\n", kwBuffer, strlen(kwBuffer), feof(f));
	
	// find corresponding keyword
	int i = 0;
	while (possibleKeywords[i].kw != KW_END_OF_KEYWORDS) {
		if (strcmp(kwBuffer, possibleKeywords[i].str) == 0)
			return possibleKeywords[i].kw;
		++i;
	}
	
	fprintf(stderr, "Unknwon keyword: %s\n", kwBuffer);
	return KW_UNKNOWN;
}


/**
 * Read the rest of the current line from file. 
 * @param f                file descriptor
 * @return string read
 */
char* readString(FILE* f) {
	// read the string
	char buffer[MAX_STRING_LENGTH + 1];
	int l = 0;
	char c = getc(f);
	while (c != '\n' && c != '\r') {
		if (l < MAX_STRING_LENGTH) {
			buffer[l] = c;
			++l;
		}
		c = getc(f);
	}
	ungetc(c, f);
	buffer[l] = '\0';
	
	// and return it
	char* str = malloc((l+1) * sizeof(char));
	strcpy(str, buffer);
	return str;
}


/**
 * Read a colon if any from file.
 */
void skipColon(FILE* f) {
	char c = getc(f);
	if (c != ':')
		ungetc(c, f);
}

/**
 * Read following whitespace characters (if any) from file.
 */
void skipWhitespace(FILE* f) {
	char c = getc(f);
	while (c == ' ' || c == '\t')
		c = getc(f);
	ungetc(c, f);
}

/**
 * Read newlines (if any) from file.
 */
void skipNewlines(FILE* f) {
	int c = getc(f);
	while (c == '\r' || c == '\n')
		c = getc(f);
	if (c != EOF)
		ungetc(c, f);
}

/**
 * Read newline from file.
 * @return true if newline was read, false otherwise
 */
bool expectNewline(FILE* f) {
	int c = getc(f);
	bool ok = false;
	if (c == '\r') {
		ok = true;
		c = getc(f);
	}
	if (c == '\n')
		ok |= true;
	else if (c != EOF)
		ungetc(c, f);
	return ok;
}

/**
 * Read a value of keyword type from file and place it in kw.
 * @param f                 file descriptor
 * @param possibleKeywords  array of allowed keywords
 * @param typeName          type string representation to be used in error messages
 * @param kw                keyword value that was read (return value)
 * @return true in success, false otherwise
 */
bool readKeywordValue(FILE* f, KeywordStr const* possibleKeywords, char const* typeName, Keyword* kw) {
	*kw = readKeyword(f, possibleKeywords);
	if (*kw == KW_UNKNOWN) {
		fprintf(stderr, "Error: Unknow %s.\n", typeName);
		return false;
	}
	skipWhitespace(f);
	if (!expectNewline(f)) {
		fprintf(stderr, "Error: Newline expected after %s.\n", typeName);
		return false;
	}
	return true;
}

/**
 * Read node coordinates from file and write them to given TSP instance.
 * @param f     file descriptor
 * @param tsp   TSP instance, that is being constructed
 * @param type  coordinate type
 * @return true in success, false otherwise
 */
bool readNodeCoordinates(FILE* f, TSP* tsp, Keyword type) {
	// get the number of dimensions
	if (type == KW_THREED_COORDS) {
		tsp->nDims = 3;
	} else {
		if (type != KW_TWOD_COORDS)
			fprintf(stderr, "Warning: Node coord type not defined, assuming 2d-coords.\n");
		tsp->nDims = 2;
	}
	
	// read coordinates
	tsp->coords = malloc(tsp->nCities * tsp->nDims * sizeof(double));
	for (int i = 0; i < tsp->nCities; ++i) {
		int j;
		
		// read city number
		if (!fscanf(f, "%d", &j)) {
			fprintf(stderr, "Error: Could not read node coordinates. (%d. coordinate line)\n", i);
			return false;
		}
		if (j <= 0 || j > tsp->nCities) {
			fprintf(stderr, "Error: City number out of range. (%d. coordinate line)\n", i);
			return false;
		}
		
		// read coords
		for (int d = 0; d < tsp->nDims; ++d) {
			double coord;
			if (!fscanf(f, "%lf", &coord)) {
				fprintf(stderr, "Error: Could not read node coordinates. (%d. coordinate line)\n", i);
				return false;
			}
			tsp->coords[(j-1) * tsp->nDims + d] = coord;
		}
		
		skipWhitespace(f);
		
		// read newline
		if (!expectNewline(f)) {
			fprintf(stderr, "Error: Newline expected. (%d. coordinate line)\n", i);
			return false;
		}
		//printf("%5d: %6.2f %6.2f\n", j, tsp->coords[(j-1) * tsp->dims + 0], tsp->coords[(j-1) * tsp->dims + 1]);
	}
	return true;	
}



/**
 * Calculate edge weights / distances between cities.
 * @param tsp             TSP instance, that is being constructed
 * @param edgeWeightType  edge weight type
 * @return true in success, false otherwise
 */
bool calculateEdgeWeights(TSP* tsp, Keyword edgeWeightType) {
	
	// number of dimensions
	int nDimsRequired = 0;
	switch (edgeWeightType) {
		case KW_EUC_2D:
		case KW_MAX_2D:
		case KW_MAN_2D:
		case KW_CEIL_2D:
		case KW_GEO:
			nDimsRequired = 2;
			break;
		case KW_EUC_3D:
		case KW_MAX_3D:
		case KW_MAN_3D:
			nDimsRequired = 3;
			break;
		case KW_ATT:
		case KW_XRAY1:
		case KW_XRAY2:
		case KW_SPECIAL:
		default:
			fprintf(stderr, "Error: Edge weight type not supported.\n");
			return false;
	}

	if (tsp->nDims != nDimsRequired) {
		fprintf(stderr, "Error: Coordinate dimension does not match with edge weight type.\n");
		return false;
	}
	
	switch (edgeWeightType) {
		// Euclidean distance
		case KW_EUC_2D:
		case KW_EUC_3D:
		case KW_CEIL_2D:
			for (int i = 0; i < tsp->nCities; ++i)
				for (int j = 0; j < tsp->nCities; ++j) {
					double ds = 0;
					for (int d = 0; d < tsp->nDims; ++d)
						ds += sq(tsp->coords[j * tsp->nDims + d] - tsp->coords[i * tsp->nDims + d]);
					tsp->dists[i * tsp->nCities + j] = sqrt(ds);
				}
			if (edgeWeightType == KW_CEIL_2D)
				for (int i = 0; i < tsp->nCities; ++i)
					for (int j = 0; j < tsp->nCities; ++j)
						tsp->dists[i * tsp->nCities + j] = ceil(tsp->dists[i * tsp->nCities + j]);
			break;
		
		// Maximum distance
		case KW_MAX_2D:
		case KW_MAX_3D:
			for (int i = 0; i < tsp->nCities; ++i)
				for (int j = 0; j < tsp->nCities; ++j) {
					double dist = 0;
					for (int d = 0; d < tsp->nDims; ++d) {
						double newDist = fabs(tsp->coords[j * tsp->nDims + d] - tsp->coords[i * tsp->nDims + d]);
						if (newDist > dist)
							dist = newDist;
					}
					tsp->dists[i * tsp->nCities + j] = dist;
				}
			break;
		
		// Manhattan distance
		case KW_MAN_2D:
		case KW_MAN_3D:
			for (int i = 0; i < tsp->nCities; ++i)
				for (int j = 0; j < tsp->nCities; ++j) {
					double dist = 0;
					for (int d = 0; d < tsp->nDims; ++d)
						dist += fabs(tsp->coords[j * tsp->nDims + d] - tsp->coords[i * tsp->nDims + d]);
					tsp->dists[i * tsp->nCities + j] = dist;
				}
			break;
		
		// Geographical distance
		case KW_GEO: {
			double latitude[tsp->nCities];
			double longitude[tsp->nCities];
			for (int i = 0; i < tsp->nCities; ++i) {
					double deg = floor(tsp->coords[i * tsp->nDims + 0]);
					double min = tsp->coords[i * tsp->nDims + 0] - deg;
					latitude[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
					deg = floor(tsp->coords[i * tsp->nDims + 1]);
					min = tsp->coords[i * tsp->nDims + 1] - deg;
					longitude[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
			}
			for (int i = 0; i < tsp->nCities; ++i)
				for (int j = 0; j < tsp->nCities; ++j) {
					double q1 = cos(longitude[j] - longitude[i]);
					double q2 = cos(latitude[j] - latitude[i]);
					double q3 = cos(latitude[j] + latitude[i]);
					tsp->dists[i * tsp->nCities + j] =
							EARTH_RADIUS * acos(0.5 * ((1 + q1) * q2 - (1 - q1) * q3)) + 1;
				}
			break; }
		
		default:;
	}
	
	for (int i = 0; i < tsp->nCities; ++i)
		for (int j = 0; j < tsp->nCities; ++j)
			tsp->dists[i * tsp->nCities + j] = round(tsp->dists[i * tsp->nCities + j]);
	
	return true;
}





/**
 * Create/allocate empty TSP.
 */
TSP* createEmptyTSP() {
	TSP* tsp = malloc(sizeof(TSP));
	return tsp;
}

/**
 * Load TSP from file.
 */
TSP* loadTSP(char const* filename) {
	// open file
	FILE* f = fopen(filename, "r");
	
	// create TSP instance
	TSP* tsp = createEmptyTSP();
	
	// initialize properties
	Keyword problemType = KW_NOT_SET;
	Keyword edgeWeightType = KW_NOT_SET;
	Keyword edgeWeightFormat = KW_NOT_SET;
	Keyword edgeDataFormat = KW_NOT_SET;
	Keyword nodeCoordType = KW_NOT_SET;
	Keyword displayDataType = KW_NOT_SET;
	//Keyword nodeCoordType = KW_NO_COORDS;
	//Keyword displayDataType = KW_NO_DISPLAY;
	
	// read file
	skipNewlines(f);
	while (!feof(f)) {
		// read keyword
		skipWhitespace(f);
		Keyword kw = readKeyword(f, mainKeywords);
		skipWhitespace(f);
		skipColon(f);
		skipWhitespace(f);
		
		switch (kw) {
			// Specification part
			
			case KW_NAME:;
				tsp->name = readString(f);
				if (!expectNewline(f)) {
					fprintf(stderr, "Error: Newline expected.\n");
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;

			case KW_TYPE:;
				if (!readKeywordValue(f, problemTypeKeywords, "problem type", &problemType)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;

			case KW_COMMENT:;
				tsp->comment = readString(f);
				if (!expectNewline(f)) {
					fprintf(stderr, "Error: Newline expected.\n");
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;

			case KW_DIMENSION:;
				if (!fscanf(f, "%d", &(tsp->nCities))) {
					fprintf(stderr, "Error: Couldn't read dimension.\n");
					fclose(f); freeTSP(tsp); return NULL;
				}
				tsp->dists = malloc(tsp->nCities * tsp->nCities * sizeof(double));
				if (!expectNewline(f)) {
					fprintf(stderr, "Error: Newline expected.\n");
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;

			case KW_CAPACITY:
				fprintf(stderr, "Error: CAPACITY not supported.\n");
				fclose(f); freeTSP(tsp); return NULL;
				break;

			case KW_EDGE_WEIGHT_TYPE:;
				if (!readKeywordValue(f, edgeWeightTypeKeywords, "edge weight type", &edgeWeightType)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;

			case KW_EDGE_WEIGHT_FORMAT:;
				if (!readKeywordValue(f, edgeWeightFormatKeywords, "edge weight format", &edgeWeightFormat)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;

			case KW_EDGE_DATA_FORMAT:;
				if (!readKeywordValue(f, edgeDataFormatKeywords, "edge data format", &edgeDataFormat)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;
	
			case KW_NODE_COORD_TYPE:;
				if (!readKeywordValue(f, nodeCoordTypeKeywords, "node coord type", &nodeCoordType)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				//if (nodeCoordType != KW_NO_COORDS && displayDataType == KW_NOT_SET)
				//	displayDataType = KW_COORD_DISPLAY;
				break;


			case KW_DISPLAY_DATA_TYPE:;
				if (!readKeywordValue(f, displayDataTypeKeywords, "display data type", &displayDataType)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;
			
			
			// Data part
			
			case KW_NODE_COORD_SECTION:;
				if (!readNodeCoordinates(f, tsp, nodeCoordType)) {
					fclose(f); freeTSP(tsp); return NULL;
				}
				break;
			
			
			// Others
			
			case KW_UNKNOWN:
				fprintf(stderr, "Error: Unknown keyword.\n");
				fclose(f); freeTSP(tsp); return NULL;
				break;

			case KW_EOF:
				expectNewline(f);
				break;
			
			default:
				// should not happen
				fprintf(stderr, "Error: readKeyword returned invalid keyword.\n");
				fclose(f); freeTSP(tsp); return NULL;
		}
		skipNewlines(f);
	}
	// end of file => close
	fclose(f);
	
	// check 
	if (problemType != KW_TSP) {
		fprintf(stderr, "Error: Data type not supported.\n");
		fclose(f); freeTSP(tsp); return NULL;
	}
	
	// calculate edge weights if not given explicitly
	if (edgeWeightType == KW_NOT_SET) {
		fprintf(stderr, "Error: Edge weight type not given.\n");
		fclose(f); freeTSP(tsp); return NULL;
	} else if (edgeWeightType == KW_EXPLICIT) {
		// not implemented => error
		fprintf(stderr, "Error: Edge weight type KW_EXPLICIT not supported.\n");
		fclose(f); freeTSP(tsp); return NULL;
	} else {
		if (tsp->coords == NULL) {
			fprintf(stderr, "Error: No coordinates given => can't calculate edge weights!.\n");
			fclose(f); freeTSP(tsp); return NULL;
		}
		
		// calculate edge weights
		if (!calculateEdgeWeights(tsp,edgeWeightType)) {
			freeTSP(tsp); return NULL;
		}
	}
	
	return tsp;
}


/**
 * Free/deallocate given TSP.
 */
void freeTSP(TSP* tsp) {
	if (tsp->name)
		free(tsp->name);
	if (tsp->comment)
		free(tsp->comment);
	if (tsp->dists)
		free(tsp->dists);
	if (tsp->coords)
		free(tsp->coords);
	free(tsp);
}








