#include <SDL/SDL.h>
#include <GL/gl.h>
#include <math.h>

#include "visualize.h"

// picture marginal
const double PICTURE_MARGINAL = 0.05;

// global variables
SDL_Surface* screen = NULL;
double pictureRatio;
double* coords = NULL;
double nCities;

/**
 * Initialize visualizing system.
 * @param width   picture width
 * @param height  picture height
 * @param bpp     bytes per pixel
 * @param flags   flags to be passed to SDL
 * @return true in success, false otherwise
 */
bool initVisualizer(int width, int height, int bpp, Uint32 flags) {
	// initialize SDL
	if (screen || SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER)) {
		fprintf(stderr, "SDL initialization failed.\n");
		return false;
	}
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	
	pictureRatio = (double) width / height;
	
	// set OpenGL attributes
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
	// create visualizer window
	screen = SDL_SetVideoMode(width, height, bpp, SDL_OPENGL | flags);
	if (!screen) {
		fprintf(stderr, "Couldn't set videomode.\n");
		SDL_Quit();
		return false;
	}
	
	SDL_WM_SetCaption("Lin-Kernighan TSP solver", NULL);
	
	// takapintojen poistaminen
	//glEnable(GL_CULL_FACE);
	//glCullFace(GL_BACK);
	
	// ruuduntyhjennysvärit
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	
	// läpinäkyvyys
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	
	// varjostusmalli
	glShadeModel(GL_FLAT);
	//glShadeModel(GL_SMOOTH);
	
	// teksturointi
	//glEnable(GL_TEXTURE_2D);
	
	// perspektiivikorjaus
	//glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	// antialiasointi
	glEnable(GL_LINE_SMOOTH),
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	//glEnable(GL_POINT_SMOOTH);
	//glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	//glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	
	//glViewport(0, 0, width, height);

	return true;
}

/**
 * Quit visualizing system.
 */
void quitVisualizer() {
	if (!screen)
		return;
	if (coords) {
		free(coords);
		coords = NULL;
	}
	SDL_Quit();
}

/**
 * Wait for uset to press key (or close window button).
 * @return the keycode of key pressed (or -1 in case of close window button)
 */
int waitKey() {
    SDL_Event event;
    while (SDL_WaitEvent(&event)) {
    	if (event.type == SDL_QUIT)
    		return -1;
		if (event.type == SDL_KEYDOWN)
			return event.key.keysym.sym;
	}
	return 0;
} 

/**
 * Check if user has pressed a key (or close window button).
 * @return the keycode of key pressed (or -1 in case of close window button),
 *         or 0 if no key was pressed
 */
int keyPressed() {
    SDL_Event event;
    while (SDL_PollEvent(&event)) {
    	if (event.type == SDL_QUIT)
    		return -1;
		if (event.type == SDL_KEYDOWN)
			return event.key.keysym.sym;
	}
	return 0;
} 

/**
 * Set TSP that is going to be visualized.
 */
void setVisualizerTSP(TSP* tsp) {
	if (coords) {
		free(coords);
		coords = NULL;
	}
	
	// node boundaries
	double minx =  INFINITY;
	double miny =  INFINITY;
	double maxx = -INFINITY;
	double maxy = -INFINITY;
	for (int i = 0; i < tsp->nCities; ++i) {
		if (tsp->coords[i * tsp->nDims + 0] < minx)
			minx = tsp->coords[i * tsp->nDims + 0];
		if (tsp->coords[i * tsp->nDims + 0] > maxx)
			maxx = tsp->coords[i * tsp->nDims + 0];
		if (tsp->coords[i * tsp->nDims + 1] < miny)
			miny = tsp->coords[i * tsp->nDims + 1];
		if (tsp->coords[i * tsp->nDims + 1] > maxy)
			maxy = tsp->coords[i * tsp->nDims + 1];
	}
	
	// center of nodes
	double cx = (maxx + minx) / 2;
	double cy = (maxy + miny) / 2;
	
	// picture height
	double hx = (maxx - minx) / (pictureRatio - 2 * PICTURE_MARGINAL);
	double hy = (maxy - miny) / (1 - 2 * PICTURE_MARGINAL);
	double h = hx > hy ? hx : hy;
	//printf("fx = %5.1lf,  fy = %5.1lf\n", fx, fy);
	
	// picture "radius" (both horizontal and vertical)
	double rx = h * pictureRatio / 2;
	double ry = h / 2;
	//printf("rx = %5.1lf,  ry = %5.1lf\n", rx, ry);
	
	// initialize OpenGL projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//glScalef(1.0, -1.0, 1.0);
	glOrtho(cx-rx, cx+rx, cy-ry, cy+ry, -1.0, 1.0);
	
	// copy node coordinates
	coords = malloc(tsp->nCities * 2 * sizeof(double));
	nCities = tsp->nCities;
	for (int i = 0; i < tsp->nCities; ++i) {
		coords[i * 2 + 0] = tsp->coords[i * tsp->nDims + 0];
		coords[i * 2 + 1] = tsp->coords[i * tsp->nDims + 1];
	}
	
}

/**
 * Clear visualizer picture.
 */
void clearPicture() {
	glClear(GL_COLOR_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/**
 * Update visualizer picture (to the screen).
 */
void updatePicture() {
	SDL_GL_SwapBuffers();
}

/**
 * Draw all nodes.
 */
void drawNodes() {
	if (!coords) {
		fprintf(stderr, "drawTSP: no coordinates => call setTSP\n");
		return;
	}
	
	glPointSize(6.0);
	glColor3f(0.0f, 0.0f, 0.0f);
	glBegin(GL_POINTS);
	for (int i = 0; i < nCities; ++i) {
		glVertex2f(coords[i * 2 + 0], coords[i * 2 + 1]);
	}
	glEnd();
	
}

void highlightNode(int i) {
	glPointSize(16.0);
	glColor4f(1.0f, 0.0f, 0.0f, 0.5f);
	glBegin(GL_POINTS);
		glVertex2f(coords[i * 2 + 0], coords[i * 2 + 1]);
	glEnd();
}

void highlightEdge(int i, int j) {
	glLineWidth(4.0);
	glColor4f(0.0f, 1.0f, 0.0f, 0.5f);
	glBegin(GL_LINES);
		glVertex2f(coords[i * 2 + 0], coords[i * 2 + 1]);
		glVertex2f(coords[j * 2 + 0], coords[j * 2 + 1]);
	glEnd();
}

void highlightForwardEdge(Route* route, int i) {
	glLineWidth(4.0);
	glColor4f(0.0f, 1.0f, 0.0f, 0.5f);
	int j = route->order[(route->position[i] + 1) % route->n];
	glBegin(GL_LINES);
		glVertex2f(coords[i * 2 + 0], coords[i * 2 + 1]);
		glVertex2f(coords[j * 2 + 0], coords[j * 2 + 1]);
	glEnd();
}

/**
 * Draw all edges.
 */
void drawEdges() {	
	glLineWidth(1.5);
	glColor3f(0.7f, 0.7f, 0.7f);
	//glColor3f(0.0f, 0.0f, 0.0f);
	glBegin(GL_LINES);
	for (int i = 0; i < nCities; ++i)
		for (int j = i+1; j < nCities; ++j) {
		glVertex2f(coords[i * 2 + 0], coords[i * 2 + 1]);
		glVertex2f(coords[j * 2 + 0], coords[j * 2 + 1]);
		}
	glEnd();	
}

/**
 * Draw edges on given route.
 */
void drawRoute(Route* route) {	
	glLineWidth(2.0);
	glColor3f(0.0f, 0.0f, 1.0f);
	glBegin(GL_LINE_LOOP);
	for (int i = 0; i < nCities; ++i) {
		int j = route->order[i];
		glVertex2f(coords[j * 2 + 0], coords[j * 2 + 1]);
	}
	glEnd();	
}




